@@ -102,6 +102,9 @@ class Options:  # pylint: disable=too-many-instance-attributes
     start_new_session: bool = False
     "True if child process should start a new session with setsid()."
 
+    preexec_fn: Any = None
+    "Function to call in child process after fork from parent."
+
     pty: bool = False
     "True if child process should be controlled using a pseudo-terminal (pty)."
 
@@ -215,6 +218,7 @@ class CmdContext:
         pass_fds_closed=_UNSET,
         write_mode=_UNSET,
         start_new_session=_UNSET,
+        preexec_fn=_UNSET,
         pty=_UNSET,
     ) -> "CmdContext":
         "Return new context with custom options set."
@@ -350,6 +354,7 @@ class Command:
         pass_fds_close: Unset[bool] = _UNSET,
         write_mode: Unset[bool] = _UNSET,
         start_new_session: Unset[bool] = _UNSET,
+        preexec_fn: Unset[Any] = _UNSET,
         pty: Unset[bool] = _UNSET,
     ) -> "Command":
         """Return new command with custom options set.
@@ -375,9 +380,11 @@ class Command:
         - Set `pass_fds_close` to True to auto-close the `pass_fds`.
         - Set `write_mode` to True when using process substitution for writing.
         - Set `start_new_session` to True to start a new session.
+        - Set `preexec_fn` to a function to call in child process.
         - Set `pty` to True to use a pseudo-terminal (pty) to control the child
-        process. This setting forces `start_new_session` to
-        also be True.
+        process. You may also set `pty` to a 1-arg function to call on the
+        child_fd for setup purposes. Setting `pty` forces `start_new_session`
+        to True.
         """
         kwargs = locals()
         del kwargs["self"]
