@@ -4,11 +4,16 @@ import asyncio
 
 from shellous.log import LOGGER
 
-# FIXME: Pass the cancel_timeout into harvest_results as an arg.
-_CANCEL_TIMEOUT = 10.0  # seconds to wait for cancelled task to finish
+_CANCEL_TIMEOUT = 15.0  # seconds to wait for cancelled task to finish
 
 
-async def harvest(*aws, timeout=None, trustee=None):
+async def harvest(
+    *aws,
+    timeout=None,
+    cancel_timeout=_CANCEL_TIMEOUT,
+    trustee=None,
+    cancel_finish=False,
+):
     """Run a bunch of awaitables as tasks. Do not return results.
 
     After the harvest returns, all of the awaitables are guaranteed to be done.
@@ -24,18 +29,30 @@ async def harvest(*aws, timeout=None, trustee=None):
     `asyncio.TimeoutError`.
 
     If `harvest` is cancelled itself, all awaitables are cancelled and
-    consumed before raising `CancelledError`.
+    consumed before raising `CancelledError`. If `cancel_finish` is True, the
+    tasks are not cancelled, but allowed to finish.
     """
 
     tasks = [asyncio.ensure_future(item) for item in aws]
-    await harvest_wait(tasks, timeout=timeout, trustee=trustee)
+    await harvest_wait(
+        tasks,
+        timeout=timeout,
+        cancel_timeout=cancel_timeout,
+        cancel_finish=cancel_finish,
+        trustee=trustee,
+    )
     _consume_exceptions(tasks)
     for task in tasks:
         if not task.cancelled():
             task.result()
 
 
-async def harvest_results(*aws, timeout=None, trustee=None):
+async def harvest_results(
+    *aws,
+    timeout=None,
+    cancel_timeout=_CANCEL_TIMEOUT,
+    trustee=None,
+):
     """Run a bunch of awaitables as tasks and return (cancelled, results).
 
     ```
@@ -61,13 +78,25 @@ async def harvest_results(*aws, timeout=None, trustee=None):
     tasks = [asyncio.ensure_future(item) for item in aws]
     cancelled = False
     try:
-        await harvest_wait(tasks, timeout=timeout, trustee=trustee)
+        await harvest_wait(
+            tasks,
+            timeout=timeout,
+            cancel_timeout=cancel_timeout,
+            trustee=trustee,
+        )
     except asyncio.CancelledError:
         cancelled = True
     return cancelled, [_to_result(task) for task in tasks]
 
 
-async def harvest_wait(tasks, *, timeout=None, trustee=None):
+async def harvest_wait(
+    tasks,
+    *,
+    timeout=None,
+    cancel_timeout=_CANCEL_TIMEOUT,
+    cancel_finish=False,
+    trustee=None,
+):
     """Wait for tasks to finish or raise an exception.
 
     After the harvest returns, all of the tasks are guaranteed to be done.
@@ -80,7 +109,8 @@ async def harvest_wait(tasks, *, timeout=None, trustee=None):
     `asyncio.TimeoutError`.
 
     If `harvest_wait` is cancelled itself, all tasks are cancelled and
-    consumed before raising `CancelledError`.
+    consumed before raising `CancelledError`. If `cancel_finish` is True,
+    the tasks are not cancelled, but allowed to finish.
     """
 
     try:
@@ -102,13 +132,17 @@ async def harvest_wait(tasks, *, timeout=None, trustee=None):
 
     except asyncio.CancelledError:
         # Cancel all tasks and wait for them to finish.
-        LOGGER.info("harvest_wait cancelled trustee=%r", trustee)
-        await _cancel_wait(tasks, trustee)
+        LOGGER.info(
+            "harvest_wait cancelled trustee=%r cancel_finish=%r",
+            trustee,
+            cancel_finish,
+        )
+        await _cancel_wait(tasks, trustee, cancel_timeout, cancel_finish)
         _consume_exceptions(tasks)
         raise
 
     if pending:
-        await _cancel_wait(pending, trustee)
+        await _cancel_wait(pending, trustee, cancel_timeout)
 
     assert all(task.done() for task in tasks)
 
@@ -118,15 +152,17 @@ async def harvest_wait(tasks, *, timeout=None, trustee=None):
         raise asyncio.TimeoutError()
 
 
-async def _cancel_wait(tasks, trustee):
+async def _cancel_wait(tasks, trustee, cancel_timeout, cancel_finish=False):
     "Cancel tasks and wait for them to finish."
     try:
-        for task in tasks:
-            task.cancel()
+        if not cancel_finish:
+            # Cancel all tasks.
+            for task in tasks:
+                task.cancel()
 
         _, pending = await asyncio.wait(
             tasks,
-            timeout=_CANCEL_TIMEOUT,
+            timeout=cancel_timeout,
             return_when=asyncio.ALL_COMPLETED,
         )
 
