@@ -9,18 +9,17 @@ import shellous
 import shellous.redirect as redir
 from shellous import pty_util
 from shellous.harvest import harvest, harvest_results
-from shellous.log import LOGGER, log_method, log_timer
+from shellous.log import LOG_DETAIL, LOG_ENTER, LOG_EXIT, LOGGER, log_method, log_timer
 from shellous.redirect import Redirect
 from shellous.result import Result, make_result
-from shellous.util import close_fds
-
-_NORMAL_LOGGING = True
-_DETAILED_LOGGING = True
+from shellous.util import close_fds, verify_dev_fd
 
 _KILL_TIMEOUT = 3.0
 _CLOSE_TIMEOUT = 0.25
 
 _KILL_EXIT_CODE = -9 if sys.platform != "win32" else 1
+_FLAKY_EXIT_CODE = 255
+_BSD = sys.platform.startswith("freebsd") or sys.platform == "darwin"
 
 
 def _is_cancelled(ex):
@@ -118,6 +117,9 @@ class _RunOptions:
             pass_fds_close=True,
         )
 
+        if _BSD:
+            verify_dev_fd(pass_fds[0])
+
     def _setup_redirects(self):
         "Set up I/O redirections."
         options = self.command.options
@@ -142,8 +144,9 @@ class _RunOptions:
 
         # Set up PTY here. This is the first half. Second half in `Runner`.
         start_new_session = options.start_new_session
-        preexec_fn = None
+        preexec_fn = options.preexec_fn
         if options.pty:
+            assert preexec_fn is None
             stdin, stdout, stderr, preexec_fn = self._setup_pty1(
                 stdin,
                 stdout,
@@ -239,7 +242,10 @@ class _RunOptions:
         Initializes `self.pty_fds`.
         """
 
-        parent_fd, child_fd = pty_util.open_pty()
+        self.pty_fds, child_fd = pty_util.open_pty(pty)
+        self.open_fds.append(child_fd)
+
+        LOGGER.info("_setup_pty1: %r child_fd=%r", self.pty_fds, child_fd)
 
         if stdin == asyncio.subprocess.PIPE:
             stdin = child_fd
@@ -252,19 +258,6 @@ class _RunOptions:
         elif stderr == asyncio.subprocess.PIPE:
             raise RuntimeError("pty can't separate stderr from stdout")
 
-        # If pty is a callable, call it here with `child_fd` as argument. This
-        # gives the client an opportunity to configure the tty.
-        if callable(pty):
-            pty(child_fd)
-
-        self.pty_fds = pty_util.PtyFds(
-            parent_fd,
-            child_fd,
-            pty_util.get_eof(child_fd),
-        )
-
-        LOGGER.info("_setup_pty1: %r", self.pty_fds)
-
         return stdin, stdout, stderr, lambda: pty_util.set_ctty(child_fd)
 
 
@@ -326,13 +319,16 @@ class Runner:
         self.tasks.append(task)
         return task
 
-    @log_method(_DETAILED_LOGGING)
+    @log_method(LOG_DETAIL)
     async def _wait(self):
         "Normal wait for background I/O tasks and process to finish."
         assert self.proc
 
         try:
-            await harvest(*self.tasks, trustee=self)
+            if self.tasks:
+                await harvest(*self.tasks, trustee=self)
+            if _BSD and self.options.pty_fds:
+                await self._wait_pid()
 
         except asyncio.CancelledError:
             LOGGER.info("Runner.wait cancelled %r", self)
@@ -340,7 +336,24 @@ class Runner:
             self.tasks.clear()  # all tasks were cancelled
             await self._kill()
 
-    @log_method(_DETAILED_LOGGING)
+    @log_method(LOG_DETAIL)
+    async def _wait_pid(self):
+        "Manually poll `waitpid` until process finishes."
+        proc_pid = self.proc.pid
+        assert proc_pid
+
+        while True:
+            pid, status = os.waitpid(proc_pid, os.WNOHANG)
+            LOGGER.info("waitpid returned %r", (pid, status))
+
+            if pid == proc_pid:
+                self.proc._transport._returncode = status
+                self.proc._transport._proc.returncode = status
+                break
+
+            await asyncio.sleep(0.025)
+
+    @log_method(LOG_DETAIL)
     async def _kill(self):
         "Kill process and wait for it to finish."
         assert self.proc
@@ -380,7 +393,7 @@ class Runner:
         else:
             self.proc.send_signal(sig)
 
-    @log_method(_DETAILED_LOGGING)
+    @log_method(LOG_DETAIL)
     async def _kill_wait(self):
         "Wait for killed process to exit."
         assert self.proc
@@ -396,7 +409,7 @@ class Runner:
             LOGGER.error("%r failed to kill process %r", self, self.proc)
             raise RuntimeError(f"Unable to kill process {self.proc!r}") from ex
 
-    @log_method(_NORMAL_LOGGING, _info=True)
+    @log_method(LOG_ENTER, _info=True)
     async def __aenter__(self):
         "Set up redirections and launch subprocess."
         try:
@@ -406,7 +419,7 @@ class Runner:
                 # Raises ResultError instead of CancelledError.
                 self.result()
 
-    @log_method(_DETAILED_LOGGING)
+    @log_method(LOG_DETAIL)
     async def _start(self):
         "Set up redirections and launch subprocess."
         assert self.proc is None
@@ -415,25 +428,15 @@ class Runner:
         try:
             # Set up subprocess arguments and launch subprocess.
             with self.options as opts:
-                # Launch the main subprocess.
-                with log_timer("asyncio.create_subprocess_exec"):
-                    self.proc = await asyncio.create_subprocess_exec(
-                        *opts.args,
-                        **opts.kwd_args,
-                    )
-
-                # Launch the process substitution commands (if any).
-                for cmd in opts.subcmds:
-                    self.add_task(cmd.coro(), "procsub")
+                await self._subprocess_exec(opts)
 
             stdin = self.proc.stdin
             stdout = self.proc.stdout
             stderr = self.proc.stderr
 
-            # Second half of pty setup.
+            # Assign pty streams.
             if opts.pty_fds:
                 assert (stdin, stdout, stderr) == (None, None, None)
-                opts.pty_fds = await opts.pty_fds.open_streams()
                 stdin, stdout = opts.pty_fds.writer, opts.pty_fds.reader
 
             if stderr is not None:
@@ -456,7 +459,7 @@ class Runner:
                     stdin = None
 
         except (Exception, asyncio.CancelledError) as ex:
-            LOGGER.info("Runner.start %r ex=%r", self, ex)
+            LOGGER.info("Runner._start %r ex=%r", self, ex)
             if _is_cancelled(ex):
                 self.cancelled = True
             if self.proc:
@@ -470,16 +473,35 @@ class Runner:
         self.stderr = stderr
 
         # Add a task to monitor for when the process finishes.
-        self.add_task(self._waiter(), "waiter")
+        if not (_BSD and opts.pty_fds):
+            self.add_task(self._waiter(), "waiter")
 
         return self
 
-    @log_method(_DETAILED_LOGGING)
+    @log_method(LOG_DETAIL)
+    async def _subprocess_exec(self, opts):
+        "Start the subprocess and assign to `self.proc`."
+
+        # Second half of pty setup.
+        if opts.pty_fds:
+            opts.pty_fds = await opts.pty_fds.open_streams()
+            if _BSD:
+                pty_util.patch_child_watcher()
+
+        with log_timer("asyncio.create_subprocess_exec"):
+            self.proc = await asyncio.create_subprocess_exec(
+                *opts.args,
+                **opts.kwd_args,
+            )
+
+        # Launch the process substitution commands (if any).
+        for cmd in opts.subcmds:
+            self.add_task(cmd.coro(), "procsub")
+
+    @log_method(LOG_DETAIL)
     async def _waiter(self):
         "Run task that waits for process to exit."
         await self.proc.wait()
-        if self.options.pty_fds:
-            self.stdout._transport.close()  # pylint: disable=protected-access
 
     def _setup_output_sink(self, stream, sink, encoding, tag):
         "Set up a task to write to custom output sink."
@@ -496,7 +518,7 @@ class Runner:
             stream = None
         return stream
 
-    @log_method(_NORMAL_LOGGING, exc_value=2)
+    @log_method(LOG_EXIT, exc_value=2)
     async def __aexit__(self, _exc_type, exc_value, _exc_tb):
         "Wait for process to exit and handle cancellation."
         suppress = False
@@ -507,7 +529,7 @@ class Runner:
             self.cancelled = True
         return suppress
 
-    @log_method(_DETAILED_LOGGING)
+    @log_method(LOG_DETAIL)
     async def _finish(self, exc_value):
         "Finish the run. Return True only if `exc_value` should be suppressed."
         assert self.proc
@@ -525,7 +547,7 @@ class Runner:
         finally:
             await self._close()
 
-    @log_method(_DETAILED_LOGGING)
+    @log_method(LOG_DETAIL)
     async def _close(self):
         "Make sure that our resources are properly closed."
         assert self.proc
@@ -541,6 +563,10 @@ class Runner:
             self.proc._transport.close()  # pylint: disable=protected-access
             return
 
+        # asyncio child watcher artifact.
+        if self.proc.returncode == _FLAKY_EXIT_CODE:
+            LOGGER.warning("Runner._close exit code=%r", self.proc.returncode)
+
         try:
             # Make sure the transport is closed (for asyncio and uvloop).
             self.proc._transport.close()  # pylint: disable=protected-access
@@ -651,7 +677,7 @@ class PipeRunner:  # pylint: disable=too-many-instance-attributes
         self.tasks.append(task)
         return task
 
-    @log_method(_DETAILED_LOGGING)
+    @log_method(LOG_DETAIL)
     async def _wait(self, *, kill=False):
         "Wait for pipeline to finish."
         assert self.results is None
@@ -666,7 +692,7 @@ class PipeRunner:  # pylint: disable=too-many-instance-attributes
         if cancelled:
             self.cancelled = True
 
-    @log_method(_NORMAL_LOGGING)
+    @log_method(LOG_ENTER)
     async def __aenter__(self):
         "Set up redirections and launch pipeline."
         try:
@@ -678,7 +704,7 @@ class PipeRunner:  # pylint: disable=too-many-instance-attributes
             await self._wait(kill=True)  # FIXME
             raise
 
-    @log_method(_NORMAL_LOGGING, exc_value=2)
+    @log_method(LOG_EXIT, exc_value=2)
     async def __aexit__(self, _exc_type, exc_value, _exc_tb):
         "Wait for pipeline to exit and handle cancellation."
         suppress = False
@@ -689,7 +715,7 @@ class PipeRunner:  # pylint: disable=too-many-instance-attributes
             self.cancelled = True
         return suppress
 
-    @log_method(_DETAILED_LOGGING)
+    @log_method(LOG_DETAIL)
     async def _finish(self, exc_value):
         "Wait for pipeline to exit and handle cancellation."
         if exc_value is not None:
@@ -700,7 +726,7 @@ class PipeRunner:  # pylint: disable=too-many-instance-attributes
 
         await self._wait()
 
-    @log_method(_DETAILED_LOGGING)
+    @log_method(LOG_DETAIL)
     async def _start(self):
         "Set up redirection and launch pipeline."
         open_fds = []
@@ -750,7 +776,7 @@ class PipeRunner:  # pylint: disable=too-many-instance-attributes
 
         return cmds
 
-    @log_method(_DETAILED_LOGGING)
+    @log_method(LOG_DETAIL)
     async def _setup_capturing(self, cmds):
         """Set up capturing and return (stdin, stdout, stderr) streams."""
 
